/*
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at <http://www.mozilla.org/MPL/>.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, WITHOUT
 * WARRANTY OF ANY KIND, either express or implied. See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * The Original Code is the Venice Web Communities System.
 * 
 * The Initial Developer of the Original Code is Eric J. Bowersox <erbo@silcom.com>,
 * for Silverwrist Design Studios.  Portions created by Eric J. Bowersox are
 * Copyright (C) 2003 Eric J. Bowersox/Silverwrist Design Studios.  All Rights Reserved.
 * 
 * Contributor(s): 
 */
package com.silverwrist.venice.community;

import java.security.Principal;
import java.security.acl.AclNotFoundException;
import java.util.*;
import org.apache.commons.collections.*;
import org.apache.log4j.Logger;
import com.silverwrist.dynamo.db.NamespaceCache;
import com.silverwrist.dynamo.db.UserManagement;
import com.silverwrist.dynamo.iface.*;
import com.silverwrist.dynamo.except.*;
import com.silverwrist.dynamo.security.SecurityReferenceMonitor;
import com.silverwrist.dynamo.util.*;
import com.silverwrist.venice.CommunityVisibility;
import com.silverwrist.venice.VeniceNamespaces;
import com.silverwrist.venice.event.*;
import com.silverwrist.venice.iface.*;

class CommunityImpl implements VeniceCommunity
{
  /*--------------------------------------------------------------------------------
   * Static data members
   *--------------------------------------------------------------------------------
   */

  private static Logger logger = Logger.getLogger(CommunityImpl.class);

  /*--------------------------------------------------------------------------------
   * Attributes
   *--------------------------------------------------------------------------------
   */

  private CommunityOps m_ops;               // community operations object
  private NamespaceCache m_nscache;         // namespace cache object
  private SecurityReferenceMonitor m_srm;   // security reference monitor
  private UserManagement m_users;           // user management object
  private AuthenticatorLookup m_alook;      // authenticator lookup object
  private PostDynamicUpdate m_post;         // where we post dynamic update events
  private CategoryService m_cats;           // category service object
  private int m_id;                         // community ID
  private int m_member_gid;                 // member group ID
  private int m_host_gid;                   // hosts group ID
  private int m_host_uid;                   // primary host UID
  private int m_aclid;                      // ID of the community ACL
  private int m_catid;                      // category ID for the community
  private CommunityVisibility m_visibility; // community visibility
  private String m_name;                    // community name
  private String m_alias;                   // community alias
  private java.util.Date m_created;         // date created
  private java.util.Date m_lastaccessed;    // date last accessed
  private java.util.Date m_lastupdate;      // date last updated
  private ReferenceMap m_properties;        // properties cache

  /*--------------------------------------------------------------------------------
   * Constructors
   *--------------------------------------------------------------------------------
   */

  CommunityImpl(CommunityOps ops, NamespaceCache nscache, SecurityReferenceMonitor srm, UserManagement users,
		AuthenticatorLookup alook, PostDynamicUpdate post, CategoryService cats, Map data)
  {
    m_ops = ops;
    m_nscache = nscache;
    m_srm = srm;
    m_users = users;
    m_alook = alook;
    m_post = post;
    m_cats = cats;
    m_id = ((Integer)(data.get(CommunityManagerOps.KEY_CID))).intValue();
    m_member_gid = ((Integer)(data.get(CommunityManagerOps.KEY_MEMBER_GID))).intValue();
    m_host_gid = ((Integer)(data.get(CommunityManagerOps.KEY_HOST_GID))).intValue();
    m_host_uid = ((Integer)(data.get(CommunityManagerOps.KEY_HOST_UID))).intValue();
    m_aclid = ((Integer)(data.get(CommunityManagerOps.KEY_ACLID))).intValue();
    m_catid = ((Integer)(data.get(CommunityManagerOps.KEY_CATID))).intValue();
    m_visibility = (CommunityVisibility)(data.get(CommunityManagerOps.KEY_VISIBILITY));
    m_name = (String)(data.get(CommunityManagerOps.KEY_NAME));
    m_alias = (String)(data.get(CommunityManagerOps.KEY_ALIAS));
    m_created = (java.util.Date)(data.get(CommunityManagerOps.KEY_CREATE));
    m_lastaccessed = (java.util.Date)(data.get(CommunityManagerOps.KEY_ACCESS));
    m_lastupdate = (java.util.Date)(data.get(CommunityManagerOps.KEY_UPDATE));
    m_properties = new ReferenceMap(ReferenceMap.HARD,ReferenceMap.SOFT);

  } // end constructor

  /*--------------------------------------------------------------------------------
   * Internal operations
   *--------------------------------------------------------------------------------
   */

  private final void testPermission(DynamoUser caller, String perm_namespace, String perm_name,
				    String fail_message) throws DatabaseException, DynamoSecurityException
  {
    try
    { // perform security tests...
      if (caller.equals(m_srm.getAdminUser()))
	return;  // Administrator can do anything
      if (m_srm.getAcl(m_aclid).testPermission(caller,perm_namespace,perm_name))
	return;  // we have the right permission in the community ACL
      if (m_srm.getGlobalAcl().testPermission(caller,perm_namespace,perm_name))
	return;  // or we can be authorized by the global ACL

    } // end try
    catch (AclNotFoundException e)
    { // just fall through and throw the exception
    } // end catch

    // throw the DynamoSecurityException
    DynamoSecurityException d = new DynamoSecurityException(CommunityImpl.class,"CommunityMessages",fail_message);
    d.setParameter(0,m_name);
    throw d;

  } // end testPermission

  /*--------------------------------------------------------------------------------
   * Overrides from class Object
   *--------------------------------------------------------------------------------
   */

  public boolean equals(Object o)
  {
    if (o==null)
      return false;
    if (o instanceof VeniceCommunity)
      return (((VeniceCommunity)o).getCID()==m_id);
    return false;

  } // end equals

  public int hashCode()
  {
    return m_id;

  } // end hashCode

  public String toString()
  {
    return "Community \"" + m_name + "\"";

  } // end toString

  /*--------------------------------------------------------------------------------
   * Implementations from interface NamedObject
   *--------------------------------------------------------------------------------
   */

  /**
   * Returns the name of this community.
   *
   * @return The name of this community.
   */
  public String getName()
  {
    return m_name;

  } // end getName

  /*--------------------------------------------------------------------------------
   * Implementations from interface ObjectProvider
   *--------------------------------------------------------------------------------
   */

  /**
   * Retrieves an object from this <CODE>ObjectProvider</CODE>.
   *
   * @param namespace The namespace to interpret the name relative to.
   * @param name The name of the object to be retrieved.
   * @return The object reference specified.
   */
  public Object getObject(String namespace, String name)
  {
    if (logger.isDebugEnabled())
      logger.debug("CommunityImpl.getObject: namespace = " + namespace + ", name = " + name);
    try
    { // convert the namespace name to an ID here
      PropertyKey key = new PropertyKey(m_nscache.namespaceNameToId(namespace),name);
      Object rc = null;
      synchronized (this)
      { // start by looking in the properties map
	rc = m_properties.get(key);
	if (rc==null)
	{ // no use - need to try the database
	  rc = m_ops.getProperty(m_id,key);
	  if (rc!=null)
          { // found in the database
	    m_properties.put(key,rc);
	    logger.debug("value found in database");

	  } // end if

	} // end if
	else
	  logger.debug("value found in cache");

      } // end synchronized block

      if (rc==null)
      { // the object was not found
	logger.debug("value not found");
	throw new NoSuchObjectException(this.toString(),namespace,name);

      } // end if

      return rc;

    } // end try
    catch (DatabaseException e)
    { // translate into our NoSuchObjectException but retain the DatabaseException
      logger.debug("Database exception while doing find",e);
      throw new NoSuchObjectException(this.toString(),namespace,name,e);

    } // end catch

  } // end getObject

  /*--------------------------------------------------------------------------------
   * Implementations from interface SecureObjectStore
   *--------------------------------------------------------------------------------
   */

  public Object setObject(DynamoUser caller, String namespace, String name, Object value)
      throws DatabaseException, DynamoSecurityException
  {
    testPermission(caller,namespace,"set.property","auth.setProperty");
    DateObjectPair rc = null;

    // convert the namespace name to an ID here
    PropertyKey key = new PropertyKey(m_nscache.namespaceNameToId(namespace),name);
    synchronized (this)
    { // start by setting the database value
      rc = m_ops.setProperty(m_id,key,value);
      
      // and cache it, too
      m_properties.put(key,value);

      // and save off the update date/time
      m_lastaccessed = m_lastupdate = rc.getDate();

    } // end synchronized block

    m_post.postUpdate(new CommunityPropertyUpdateEvent(this,namespace,name));
    return rc.getObject();

  } // end setObject

  public Object removeObject(DynamoUser caller, String namespace, String name)
      throws DatabaseException, DynamoSecurityException
  {
    testPermission(caller,namespace,"remove.property","sec.removeGlobalProperty");
    DateObjectPair rc = null;

    // convert the namespace name to an ID here
    PropertyKey key = new PropertyKey(m_nscache.namespaceNameToId(namespace),name);
    synchronized (this)
    { // start by killing the database value
      rc = m_ops.removeProperty(m_id,key);

      // and remove the cached value, too
      m_properties.remove(key);

      if (rc!=null)  // save off update date.time
	m_lastaccessed = m_lastupdate = rc.getDate();

    } // end synchronized block

    m_post.postUpdate(new CommunityPropertyUpdateEvent(this,namespace,name));
    return (rc==null) ? null : rc.getObject();

  } // end removeObject

  public Collection getNamespaces() throws DatabaseException
  {
    // call through to the database to get the list of namespace IDs
    int[] ids = m_ops.getPropertyNamespaceIDs(m_id);

    ArrayList rc = new ArrayList(ids.length);
    for (int i=0; i<ids.length; i++)
      rc.add(m_nscache.namespaceIdToName(ids[i]));
    return Collections.unmodifiableList(rc);

  } // end getNamespaces

  public Collection getNamesForNamespace(String namespace) throws DatabaseException
  {
    // call through to the database to get the data for this namespace
    int nsid = m_nscache.namespaceNameToId(namespace);
    Map data = m_ops.getAllProperties(m_id,nsid);

    // we both create the return value and cache the data values
    ArrayList rc = new ArrayList(data.size());
    synchronized (this)
    { // do the transfer...
      Iterator it = data.entrySet().iterator();
      while (it.hasNext())
      { // copy one entry at a time
	Map.Entry ntry = (Map.Entry)(it.next());
	rc.add(ntry.getKey().toString());
	m_properties.put(new PropertyKey(nsid,ntry.getKey().toString()),ntry.getValue());

      } // end while

    } // end synchronized block

    return Collections.unmodifiableList(rc);

  } // end getNamesForNamespace

  /*--------------------------------------------------------------------------------
   * Implementations from interface VeniceCommunity
   *--------------------------------------------------------------------------------
   */

  public int getCID()
  {
    return m_id;

  } // end getCID

  public int getMemberGID()
  {
    return m_member_gid;

  } // end getMemberGID

  public DynamoGroup getMemberGroup() throws DatabaseException
  {
    return m_users.getGroup(m_member_gid);

  } // end getMemberGroup

  public int getHostGID()
  {
    return m_host_gid;

  } // end getHostGID

  public DynamoGroup getHostGroup() throws DatabaseException
  {
    return m_users.getGroup(m_host_gid);

  } // end getHostGroup

  public int getHostUID()
  {
    return m_host_uid;

  } // end getHostUID

  public DynamoUser getHostUser() throws DatabaseException
  {
    return m_users.getUser(m_host_uid);

  } // end getHostUser

  public DynamoAcl getAcl() throws DatabaseException, AclNotFoundException
  {
    return m_srm.getAcl(m_aclid);

  } // end getAcl

  public int getCategoryID()
  {
    return m_catid;

  } // end getCategoryID

  public VeniceCategory getCategory() throws DatabaseException
  {
    return m_cats.getCategory(m_catid);

  } // end getCategory

  public synchronized void setCategoryID(DynamoUser caller, int catid)
      throws DatabaseException, DynamoSecurityException
  {
    testPermission(caller,VeniceNamespaces.COMMUNITY_PROFILE_NAMESPACE,"set.category","auth.setCategory");
    java.util.Date update = m_ops.setCategoryID(m_id,catid);
    m_catid = catid;
    m_lastaccessed = m_lastupdate = update;
    m_post.postUpdate(new CommunityBaseUpdateEvent(this,"category"));

  } // end setCategoryID

  public void setCategory(DynamoUser caller, VeniceCategory cat) throws DatabaseException, DynamoSecurityException
  {
    this.setCategoryID(caller,cat.getCategoryID());

  } // end setCategory

  public CommunityVisibility getVisibility()
  {
    return m_visibility;

  } // end getVisibility

  public synchronized void setVisibility(DynamoUser caller, CommunityVisibility vis)
      throws DatabaseException, DynamoSecurityException
  {
    testPermission(caller,VeniceNamespaces.COMMUNITY_PROFILE_NAMESPACE,"set.visibility","auth.setVisibility");
    java.util.Date update = m_ops.setVisibility(m_id,vis);
    m_visibility = vis;
    m_lastaccessed = m_lastupdate = update;
    m_post.postUpdate(new CommunityBaseUpdateEvent(this,"visibility"));

  } // end setVisibility

  public synchronized void setName(DynamoUser caller, String name) throws DatabaseException, DynamoSecurityException
  {
    testPermission(caller,VeniceNamespaces.COMMUNITY_PROFILE_NAMESPACE,"set.name","auth.setName");
    java.util.Date update = m_ops.setName(m_id,name);
    m_name = name;
    m_lastaccessed = m_lastupdate = update;
    m_post.postUpdate(new CommunityBaseUpdateEvent(this,"name"));

  } // end setName

  public String getAlias()
  {
    return m_alias;

  } // end getAlias

  public synchronized void setAlias(DynamoUser caller, String alias) throws DatabaseException, DynamoSecurityException
  {
    testPermission(caller,VeniceNamespaces.COMMUNITY_PROFILE_NAMESPACE,"set.alias","auth.setAlias");
    java.util.Date update = m_ops.setAlias(m_id,alias);
    m_alias = alias;
    m_lastaccessed = m_lastupdate = update;
    m_post.postUpdate(new CommunityBaseUpdateEvent(this,"alias"));

  } // end setAlias

  public java.util.Date getCreationDate()
  {
    return m_created;

  } // end getCreationDate

  public java.util.Date getLastAccessDate()
  {
    return m_lastaccessed;

  } // end getLastAccessDate

  public synchronized void setLastAccessDate(DynamoUser caller, java.util.Date date) throws DatabaseException
  {
    m_ops.setLastAccessDate(m_id,date);
    m_lastaccessed = date;

  } // end setLastAccessedDate

  public java.util.Date getLastUpdateDate()
  {
    return m_lastupdate;

  } // end getLastUpdateDate

  public synchronized void setLastUpdateDate(DynamoUser caller, java.util.Date date) throws DatabaseException
  {
    m_ops.setLastUpdateDate(m_id,date);
    m_lastaccessed = m_lastupdate = date;

  } // end setLastUpdateDate

  public void grantAccess(DynamoUser caller, Principal subject, String auth_namespace, String auth_name,
			  String source_info, String auth_info, boolean single_use)
      throws AuthenticationException, DatabaseException, DynamoSecurityException
  {
    testPermission(caller,VeniceNamespaces.COMMUNITY_PERMS_NAMESPACE,"grant.revoke.access","auth.grantAccess");

    // Translate the subject into an ID and group flag.
    int ugid = -1;
    boolean is_group = false;
    if (subject instanceof DynamoUser)
      ugid = ((DynamoUser)subject).getUID();
    else if (subject instanceof DynamoGroup)
    { // it's a group
      ugid = ((DynamoGroup)subject).getGID();
      is_group = true;

    } // end else if
    else
      throw new IllegalArgumentException("invalid subject");

    // Make sure the authenticator is valid.
    int auth_nsid = 0;
    Authenticator auth = null;
    if ((auth_namespace!=null) || (auth_name!=null))
    { // look up the authenticator to make sure it's registered
      auth = m_alook.findAuthenticator(auth_namespace,auth_name);
      if (auth==null)
	throw new IllegalArgumentException("invalid authenticator");
      auth_nsid = m_nscache.namespaceNameToId(auth_namespace);

    } // end if

    // Preprocess the authentication information through the authenticator.
    String real_auth_info = auth_info;
    if ((auth_info!=null) && (auth!=null))
      real_auth_info = auth.processInputData(auth_info);

    synchronized (this)
    { // update the access table and touch the community
      java.util.Date update = m_ops.grantAccess(m_id,ugid,is_group,single_use,auth_nsid,auth_name,source_info,
						real_auth_info);
      m_lastaccessed = m_lastupdate = update;

    } // end synchronized block

  } // end grantAccess

  public void revokeAccess(DynamoUser caller, Principal subject) throws DatabaseException, DynamoSecurityException
  {
    testPermission(caller,VeniceNamespaces.COMMUNITY_PERMS_NAMESPACE,"grant.revoke.access","auth.revokeAccess");

    // Translate the subject into an ID and group flag.
    int ugid = -1;
    boolean is_group = false;
    if (subject instanceof DynamoUser)
      ugid = ((DynamoUser)subject).getUID();
    else if (subject instanceof DynamoGroup)
    { // it's a group
      ugid = ((DynamoGroup)subject).getGID();
      is_group = true;

    } // end else if
    else
      throw new IllegalArgumentException("invalid subject");

    synchronized (this)
    { // update the access table and touch the community
      java.util.Date update = m_ops.revokeAccess(m_id,ugid,is_group);
      m_lastaccessed = m_lastupdate = update;

    } // end synchronized block

  } // end revokeAccess

} // end class CommunityImpl
